import {inject, Inject, Injectable} from '@angular/core';
import {HttpClient, HttpErrorResponse} from '@angular/common/http';
import {
  ICommand,
  IDataSourceCommandAdapterOptions,
  IDataSourceError,
  IDataSourceErrorMessage,
  IDataSourceOptions,
  IDataSourceQueryAdapterOptions,
  IDataSourceTransportOptions,
  IDataSourceValidationError,
  IQueryCriteria,
  IQueryExecutionGroupResult,
  IQueryExecutionResult,
  IResolveCommandModelEvent
} from '@openharbor/data';
import {catchError, switchMap} from 'rxjs/operators';
import {Observable, of, throwError} from 'rxjs';

@Injectable({
  providedIn: 'root'
})
export class GenericRestDataSourceService
{
  protected readonly http = inject(HttpClient);

  private _handleErrorPipe(err: HttpErrorResponse) : Observable<IDataSourceError> {

    if (err.status == 500) {
      return throwError(<IDataSourceErrorMessage>{
        type: 'message',
        message: 'UNEXPECTED_ERROR_MESSAGE'
      });
    }

    if (err.status == 400)
    {
      if (err.error && err.error.errors)
        return throwError(<IDataSourceValidationError>{
          type: 'validation',
          errors: err.error.errors
        });

      // if status not okay then it's an exception error
      if (err.error.hasOwnProperty('Message') && typeof(err.error['Message']) == "string") {
        return throwError(<IDataSourceErrorMessage>{
          type: 'message',
          message: err.error['Message']
        });
      }
    }

    // general error message
    if (typeof(err.error) == "string") {
      return throwError(<IDataSourceErrorMessage>{
        type: 'message',
        message: err.error
      });
    }

    return throwError(<IDataSourceErrorMessage>{
      type: 'message',
      message: 'UNEXPECTED_ERROR_MESSAGE'
    });
  }

  createDataSourceOptions<TQuery extends IQueryCriteria, TModel extends ICommand>(route: string, keyResolver: (model: TModel) => TModel[keyof TModel], defaultCriteria: IQueryCriteria, beforeRead?: (query: IQueryCriteria) => Observable<IQueryCriteria>) : IDataSourceOptions<TQuery, TModel>
  {
    const dataSourceTransportOptions = this.createStandardRestTransportOptions<TQuery, TModel>(route, keyResolver, beforeRead);

    return {
      defaultCriteria: defaultCriteria,
      resolveIdField: keyResolver,
      transport: dataSourceTransportOptions
    };
  }

  setResolveCommand<TQuery extends IQueryCriteria, TModel>(options: IDataSourceOptions<TQuery, TModel>, name: string, resolveCommandModel: (event: IResolveCommandModelEvent<TModel>) => Observable<any>) {
    options.transport.commands[name].resolveCommandModel = resolveCommandModel;
  }

  createStandardRestTransportOptions<TQuery extends IQueryCriteria, TModel extends ICommand>(route: string, keyResolver: (model: TModel) => TModel[keyof TModel], beforeRead?: (query: IQueryCriteria) => Observable<IQueryCriteria>) : IDataSourceTransportOptions<TQuery, TModel> {
    const query: IDataSourceQueryAdapterOptions<TQuery, TModel> = {
      adapter: {
        handle: (criteria: IQueryCriteria) => {
          const queryRoute = `${route}/read`;
          const finalBeforeRead = beforeRead || (() => of(criteria));
          return finalBeforeRead(criteria)
            .pipe(switchMap(finalQuery => {
              return this.http.post<IQueryExecutionResult<TModel> & IQueryExecutionGroupResult<TModel>>(queryRoute, finalQuery);
            }));
        }
      }
    };

    const createCommand: IDataSourceCommandAdapterOptions<TModel> = {
      adapter: {
        handle: (command: TModel) => {
          return this.http.post<TModel>(route, command).pipe(catchError(this._handleErrorPipe));
        }
      }
    };

    const updateCommand: IDataSourceCommandAdapterOptions<TModel> = {
      adapter: {
        handle: (command: TModel) => {
          const key = keyResolver(command);
          const updateRoute = `${route}/${encodeURIComponent(key as any)}`;
          return this.http.put<TModel>(updateRoute, command).pipe(catchError(this._handleErrorPipe));
        }
      }
    };

    const deleteCommand: IDataSourceCommandAdapterOptions<TModel> = {
      adapter: {
        handle: (command: TModel) => {
          const key = keyResolver(command);
          const updateRoute = `${route}/${encodeURIComponent(key as any)}`;
          return this.http.delete<TModel>(updateRoute).pipe(catchError(this._handleErrorPipe));
        }
      }
    };

    return {
      query: query,
      commands: {
        'create': createCommand,
        'update': updateCommand,
        'delete': deleteCommand
      }
    };
  }
}
